#ifndef DGX_TRANSFORM_2D_H
#define DGX_TRANSFORM_2D_H

#include <engine/psx.h>

#define FLIP_H	(1<<0)
#define FLIP_V	(1<<1)

/* POLY_xx3 transformations (3 vertices) */
#define MOVE_POLY3(p, x, y) \
	(p)->x0 += x, (p)->x1 += x, (p)->x2 += x, \
	(p)->y0 += y, (p)->y1 += y, (p)->y2 += y

/* FIXME: do not use SCALE_POLY3 */
#define SCALE_POLY3(p, n) \
	(p)->x0 +=  0, (p)->x1 +=  n, (p)->x2 += -n, \
	(p)->y0 += -n

/*
 * HFLIP_POLY3 is:
 *         (x0, y0)
 * (x2, y2) <-----> (x3, y3)
 */
#define HFLIP_POLY3(p) \
{						\
	short bufx, bufy;			\
						\
	bufx = (p)->x1; bufy = (p)->y1;		\
	(p)->x1 = (p)->x2;			\
	(p)->y1 = (p)->y2;			\
	(p)->x2 = bufx;				\
	(p)->y2 = bufy;				\
}

/*
 * VFLIP_POLY3 is: (x values are unchanged)
 *          (x0, y0)      
 *              ^         
 *       _______|________ 
 *      |                |
 *      v                v
 * (x2, y2) <-----> (x3, y3)
 */
#define VFLIP_POLY3(p) \
{				\
	short bufy;		\
				\
	bufy = (p)->y0;		\
	(p)->y0 = (p)->y1;	\
	(p)->y1 = bufy;		\
	(p)->y2 = bufy;		\
}

/* POLY_xx4 transformations (4 vertices) */
#define MOVE_POLY4(p, x, y) \
	(p)->x0 += x, (p)->x1 += x, (p)->x2 += x, (p)->x3 += x, \
	(p)->y0 += y, (p)->y1 += y, (p)->y2 += y, (p)->y3 += y

#define SCALE_POLY4(p, n) \
	(p)->x0 += -n, (p)->x1 +=  n, (p)->x2 += -n, (p)->x3 += n, \
	(p)->y0 += -n, (p)->y1 += -n, (p)->y2 +=  n, (p)->y3 += n

/*
 * HFLIP_POLY4 is:
 *
 * (x0, y0) <-----> (x1, y1)
 * (x2, y2) <-----> (x3, y3)
 */
#define HFLIP_POLY4(p) \
{						\
	short bufx, bufy;			\
						\
	bufx = (p)->x0; bufy = (p)->y0;		\
	(p)->x0 = (p)->x1;			\
	(p)->y0 = (p)->y1;			\
	(p)->x1 = bufx;				\
	(p)->y1 = bufy;				\
	bufx = (p)->x2; bufy = (p)->y2;		\
	(p)->x2 = (p)->x3;			\
	(p)->y2 = (p)->y3;			\
	(p)->x3 = bufx;				\
	(p)->y3 = bufy;				\
}

/*
 * VFLIP_POLY4 is:
 *
 * (x0, y0)         (x1, y1)
 *     ^                ^
 *     |                |
 *     |                |
 *     v                v
 * (x2, y2)         (x3, y3)
 */
#define VFLIP_POLY4(p) \
{						\
	short bufx, bufy;			\
						\
	bufx = (p)->x0; bufy = (p)->y0;		\
	(p)->x0 = (p)->x2;			\
	(p)->y0 = (p)->y2;			\
	(p)->x2 = bufx;				\
	(p)->y2 = bufy;				\
	bufx = (p)->x1; bufy = (p)->y1;		\
	(p)->x1 = (p)->x3;			\
	(p)->y1 = (p)->y3;			\
	(p)->x3 = bufx;				\
	(p)->y3 = bufy;				\
}

#define ROTATE_POLY4(p, _deg, _center_x, _center_y, _is_flipped, _radius) \
{									\
	unsigned int _v0, _v1, _v2, _v3;				\
									\
	switch(_is_flipped)						\
	{								\
		case 0:							\
			_v0 = 2560;					\
			_v1 = 3584;					\
			_v2 = 1536;					\
			_v3 = 512;					\
			break;						\
		case FLIP_H:						\
			_v0 = 3584;					\
			_v1 = 2560;					\
			_v2 = 512;					\
			_v3 = 1536;					\
			break;						\
		case FLIP_V:						\
			_v0 = 1536;					\
			_v1 = 512;					\
			_v2 = 2560;					\
			_v3 = 3584;					\
			break;						\
		case FLIP_V|FLIP_H:					\
			_v0 = 512;					\
			_v1 = 1536;					\
			_v2 = 3584;					\
			_v3 = 2560;					\
			break;						\
		default:						\
			printf("Error: ROTATE_POLY4: _is_flipped has nonsensical value\n"); \
			return;						\
	}								\
									\
	_v0 = (_v0 + _deg)%4096;					\
	_v1 = (_v1 + _deg)%4096;					\
	_v2 = (_v2 + _deg)%4096;					\
	_v3 = (_v3 + _deg)%4096;					\
									\
	(p)->x0 = ((rcos(_v0) * (_radius>>12))>>12) + (_center_x>>12);	\
	(p)->y0 = ((rsin(_v0) * (_radius>>12))>>12) + (_center_y>>12);	\
									\
	(p)->x1 = ((rcos(_v1) * (_radius>>12))>>12) + (_center_x>>12);	\
	(p)->y1 = ((rsin(_v1) * (_radius>>12))>>12) + (_center_y>>12);	\
									\
	(p)->x2 = ((rcos(_v2) * (_radius>>12))>>12) + (_center_x>>12);	\
	(p)->y2 = ((rsin(_v2) * (_radius>>12))>>12) + (_center_y>>12);	\
									\
	(p)->x3 = ((rcos(_v3) * (_radius>>12))>>12) + (_center_x>>12);	\
	(p)->y3 = ((rsin(_v3) * (_radius>>12))>>12) + (_center_y>>12);	\
}

#define MOVE_LINE2(p, x, y) \
	(p)->x0 += x, (p)->y0 += y, \
	(p)->x1 += x, (p)->y1 += y

#define MOVE_LINE3 MOVE_POLY3
#define MOVE_LINE4 MOVE_POLY4

struct CharRECT
{
	char x, y;
	char w, h;
};

struct UCharRECT
{
	unsigned char x, y;
	unsigned char w, h;
};

struct ShortRECT
{
	short x, y;
	short w, h;
};

struct UShortRECT
{
	unsigned short x, y;
	unsigned short w, h;
};

struct IntRECT
{
	int x, y;
	int w, h;
};

struct UIntRECT
{
	unsigned int x, y;
	unsigned int w, h;
};

struct Vector2x
{
	short x, y;
};

struct Vector2x_uc
{
	unsigned char vx, vy;
	short pad;
};

struct Vector2u
{
	short u, v;
};

struct Vector2u_uc
{
	unsigned char u, v;
	short pad;
};

struct Vector2w
{
	short w, h;
};

#endif

